"""
usage: spetlr-test-job submit [-h] [--dry-run] [--wheels WHEELS] --tests TESTS (--task TASK | --tasks-from TASKS_FROM) (--cluster CLUSTER | --cluster-file CLUSTER_FILE)
                              [--sparklibs SPARKLIBS | --sparklibs-file SPARKLIBS_FILE] [--requirement REQUIREMENT | --requirements-file REQUIREMENTS_FILE] [--main-script MAIN_SCRIPT] [--pytest-args PYTEST_ARGS]
                              [--out-json OUT_JSON]

Run Test Cases on databricks cluster.

optional arguments:
  -h, --help            show this help message and exit
  --dry-run             Don't do anything, only report
  --wheels WHEELS       The glob paths of all wheels under test.
  --tests TESTS         Location of the tests folder. Will be sent to databricks as a whole.
  --task TASK           Single Test file or folder to execute.
  --tasks-from TASKS_FROM
                        path in test archive where each subfolder becomes a task.
  --cluster CLUSTER     JSON document describing the cluster setup.
  --cluster-file CLUSTER_FILE
                        File with JSON document describing the cluster setup.
  --sparklibs SPARKLIBS
                        JSON document describing the spark dependencies.
  --sparklibs-file SPARKLIBS_FILE
                        File with JSON document describing the spark dependencies.
  --requirement REQUIREMENT
                        a python dependency, specified like for pip
  --requirements-file REQUIREMENTS_FILE
                        File with python dependencies, specified like for pip
  --main-script MAIN_SCRIPT
                        Your own test_main.py script file, to add custom functionality.
  --pytest-args PYTEST_ARGS
                        Additional arguments to pass to pytest in each test job.
  --out-json OUT_JSON   File to store the RunID for future queries.

"""
import argparse
import copy
import datetime
import inspect
import json
import re
import shutil
import subprocess
import tempfile
import time
import uuid
from pathlib import Path
from typing import Dict, List, Union
from typing.io import IO

from . import test_main
from .dbcli import db_check, dbfscall, dbjcall
from .dbfs import DbfsLocation


def setup_submit_parser(subparsers):
    """
    Adds a subparser for the command 'submit'.
    :param subparsers: must be the object returned by ArgumentParser().add_subparsers()
    :return:
    """

    parser: argparse.ArgumentParser = subparsers.add_parser(
        "submit", description="Run Test Cases on databricks cluster."
    )
    parser.set_defaults(func=submit_main)

    parser.add_argument(
        "--dry-run", help="Don't do anything, only report", action="store_true"
    )

    parser.add_argument(
        "--wheels",
        type=str,
        required=False,
        help="The glob paths of all wheels under test.",
        default="dist/*.whl",
    )

    parser.add_argument(
        "--tests",
        type=str,
        required=True,
        help="Location of the tests folder. Will be sent to databricks as a whole.",
    )

    parser.add_argument(
        "--task", help="Single Test file or folder to execute.", action="append"
    )
    parser.add_argument(
        "--tasks-from",
        help="path in test archive where each subfolder becomes a task.",
        action="append",
    )

    # cluster argument pair
    cluster = parser.add_mutually_exclusive_group(required=True)
    cluster.add_argument(
        "--cluster",
        type=str,
        help="JSON document describing the cluster setup.",
    )
    cluster.add_argument(
        "--cluster-file",
        type=argparse.FileType("r"),
        help="File with JSON document describing the cluster setup.",
    )

    # spark libraries argument pair
    sparklibs = parser.add_mutually_exclusive_group(required=False)
    sparklibs.add_argument(
        "--sparklibs",
        help="JSON document describing the spark dependencies.",
    )
    sparklibs.add_argument(
        "--sparklibs-file",
        type=argparse.FileType("r"),
        help="File with JSON document describing the spark dependencies.",
    )

    # python dependencies file
    pydep = parser.add_mutually_exclusive_group(required=False)
    pydep.add_argument(
        "--requirement",
        action="append",
        help="a python dependency, specified like for pip",
        default=[],
    )
    pydep.add_argument(
        "--requirements-file",
        type=argparse.FileType("r"),
        help="File with python dependencies, specified like for pip",
    )

    parser.add_argument(
        "--main-script",
        type=argparse.FileType("r"),
        help="Your own test_main.py script file, to add custom functionality.",
    )

    parser.add_argument(
        "--pytest-args", help="Additional arguments to pass to pytest in each test job."
    )

    parser.add_argument(
        "--out-json",
        type=argparse.FileType("w"),
        help="File to store the RunID for future queries.",
    )

    return


def collect_arguments(args):
    """
    Post process the parsed arguments of the 'submit' command argument parser.
    :param args: parsed arguments of the 'submit' command argument parser
    :return:
    """

    # pre-process 'cluster'
    if args.cluster_file:
        args.cluster = args.cluster_file.read()
    args.cluster = json.loads(args.cluster)

    # pre-process 'sparklibs'
    if args.sparklibs_file:
        args.sparklibs = args.sparklibs_file.read()
    if args.sparklibs:
        args.sparklibs = json.loads(args.sparklibs)

    # pre-process 'requirement'
    if args.requirements_file:
        args.requirement = [
            line.strip()
            for line in args.requirements_file.read().splitlines()
            if line.strip() and not line.strip().startswith("#")
        ]

    args.pytest_args = (args.pytest_args or "").split()

    return args


def submit_main(args):
    """the main function of the cli command 'submit'. Not to be used directly."""
    db_check()

    args = collect_arguments(args)

    submit(
        test_path=args.tests,
        cluster=args.cluster,
        wheels=args.wheels,
        tasks=args.task,
        tasks_from=args.tasks_from,
        requirement=args.requirement,
        sparklibs=args.sparklibs,
        out_json=args.out_json,
        main_script=args.main_script,
        pytest_args=args.pytest_args,
        dry_run=args.dry_run,
    )


class DbTestFolder:
    """Context manager that creates a unique test folder on dbfs."""

    def __init__(self):
        self._test_path_base = DbfsLocation(
            "/".join(
                [
                    "test",
                    datetime.datetime.now(datetime.timezone.utc).strftime(
                        "%Y%m%d-%H%M%S"
                    ),
                    uuid.uuid4().hex,
                ]
            )
        )

    def __enter__(self):
        print(f"Making dbfs test folder {self._test_path_base.remote}")
        dbfscall(f"mkdirs {self._test_path_base.remote}")
        return self._test_path_base

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def verify_and_resolve_task(test_path: str, task: Union[str, Path]):
    test_archive = Path(test_path).resolve().absolute()

    task_path = Path(task)
    if not task_path.is_absolute():
        task_path = (test_archive.parent / task_path).resolve().absolute()

    task_path = task_path.resolve().absolute()

    if not (test_archive == task_path or test_archive in task_path.parents):
        raise AssertionError(f"task {task} is not contained in {test_path}")

    if not task_path.exists():
        raise AssertionError(f"task {task_path} does not exist")

    if not (test_archive.parent / task).exists():
        raise AssertionError(f"The task {task} was not found in the test location.")

    return task_path.relative_to(test_archive.parent).as_posix()


def discover_job_tasks(test_path: str, folder: str):
    """If folder is given, create parallel tasks from this level.
    Otherwise, simply return the top folder as the single task.
    Returns a list of strings with the subfolders to process.
    """

    test_archive_parent = Path(test_path).resolve().absolute().parent

    subfolders = [
        verify_and_resolve_task(test_path, x)
        for x in (test_archive_parent / folder).iterdir()
        if (x.is_dir() and not x.stem.startswith("_"))
    ]
    return subfolders


class PoolBoy:
    """Hold a list of available instance pools and replace the by-name reference with an id if possible."""

    MARKER = "instance-pool://"

    def __init__(self):
        self._lookup = self.get_instance_pools()

    def lookup(self, pool_id: str) -> str:
        if pool_id.startswith(self.MARKER):
            pool_name = pool_id[len(self.MARKER) :]
            return self._lookup[pool_name]
            # if this throws a KeyError, we are right to abort execution since the input is invalid.
        else:
            # No Marker = no lookup
            return pool_id

    def get_instance_pools(self) -> Dict[str, str]:
        pool_data = dbjcall("instance-pools list --output JSON")
        pool_lookup = {
            pool["instance_pool_name"]: pool["instance_pool_id"]
            for pool in pool_data["instance_pools"]
        }
        return pool_lookup


def submit(
    test_path: str,
    cluster: dict,
    wheels: str,
    tasks: List[str] = None,
    tasks_from: List[str] = None,
    requirement: List[str] = None,
    sparklibs: List[dict] = None,
    out_json: IO[str] = None,
    main_script: IO[str] = None,
    pytest_args: List[str] = None,
    dry_run=False,
):
    """
    --dry-run             Don't do anything, only report
    --wheels WHEELS       The glob paths of all wheels under test.
    --tests TESTS         Location of the tests folder. Will be sent to databricks as a whole.
    --task TASK           Single Test file or folder to execute. List of single tasks.
    --tasks-from TASKS_FROM
                          path in test archive where each subfolder becomes a task.
                          List of folders to split from.
    --cluster CLUSTER     JSON document describing the cluster setup.
    --cluster-file CLUSTER_FILE
                          File with JSON document describing the cluster setup.
    --sparklibs SPARKLIBS
                          JSON document describing the spark dependencies.
    --sparklibs-file SPARKLIBS_FILE
                          File with JSON document describing the spark dependencies.
    --requirement REQUIREMENT
                          a python dependency, specified like for pip
    --requirements-file REQUIREMENTS_FILE
                          File with python dependencies, specified like for pip
    --main-script MAIN_SCRIPT
                          Your own test_main.py script file, to add custom functionality.
    --pytest-args PYTEST_ARGS
                          Additional arguments to pass to pytest in each test job.
    --out-json OUT_JSON   File to store the RunID for future queries.
    """
    if requirement is None:
        requirement = []
    if sparklibs is None:
        sparklibs = []
    if pytest_args is None:
        pytest_args = []
    if tasks is None:
        tasks = []
    if tasks_from is None:
        tasks_from = []
    if not (tasks or tasks_from):
        raise ValueError("No tasks given")

    # check the structure of the cluster object
    if not isinstance(cluster, dict):
        raise AssertionError("invalid cluster specification")

    # check the structure of the sparklibs object
    if not isinstance(sparklibs, list):
        raise AssertionError("invalid sparklibs specification")

    for py_requirement in requirement:
        sparklibs.append({"pypi": {"package": py_requirement}})

    pools = PoolBoy()

    with DbTestFolder() as test_folder:
        for wheel in discover_and_push_wheels(wheels, test_folder):
            sparklibs.append({"whl": wheel})

        archive_local = archive_and_push(test_path, test_folder, dry_run=dry_run)
        main_file = push_main_file(test_folder, main_script, dry_run=dry_run)

        print(f"copied everything to {test_folder.remote}")

        if not dry_run:
            print("Wait 1 minute for uploading test folder...")
            time.sleep(60)
            print("Waited 1 minute successfully!")

        # construct the workflow object
        workflow = dict(run_name="Testing Run", format="MULTI_TASK", tasks=[])

        resolved_tasks = [verify_and_resolve_task(test_path, task) for task in tasks]
        for task in tasks_from:
            # subtasks will be ['tests/cluster/job1', 'tests/cluster/job2'] or similar
            resolved_tasks += discover_job_tasks(test_path, task)

        if dry_run:
            print(resolved_tasks)

        for task in resolved_tasks:
            task_sub = re.sub(r"[^a-zA-Z0-9_-]", "_", task)

            # prepare cluster
            task_cluster = copy.deepcopy(cluster)
            task_cluster["cluster_log_conf"] = {
                "dbfs": {"destination": f"{test_folder.remote}/{task_sub}"}
            }
            if "instance_pool_id" in task_cluster:
                task_cluster["instance_pool_id"] = pools.lookup(
                    task_cluster["instance_pool_id"]
                )

            workflow["tasks"].append(
                dict(
                    task_key=task_sub,
                    libraries=sparklibs,
                    spark_python_task=dict(
                        python_file=main_file.remote,
                        parameters=[
                            # running in the spark python interpreter, the python __file__ variable does not
                            # work. Hence, we need to tell the script where the test area is.
                            f"--basedir={test_folder.local}/{task_sub}",
                            # all tests will be unpacked from here
                            f"--archive={archive_local}",
                            # we can actually run any part of our test suite, but some files need the full repo.
                            # Only run tests from this folder.
                            f"--folder={task}",
                            # additional arguments to pass to pytest
                            f"--pytestargs={json.dumps(pytest_args)}",
                        ],
                    ),
                    new_cluster=task_cluster,
                )
            )

    print("Submitting test...")
    with tempfile.TemporaryDirectory() as tmp:
        jobfile = f"{tmp}/job.json"
        with open(jobfile, "w") as f:
            json.dump(workflow, f)

        if dry_run:
            with open(jobfile) as f:
                print("DEBUG: workflow:")
                print(f.read())

        try:
            if not dry_run:
                res = dbjcall(f"runs submit --json-file {jobfile}")
            else:
                print(f"DRY-RUN: Call: databricks runs submit --json-file {jobfile}")
                print("DRY-RUN COMPLETED")
                return
        except subprocess.CalledProcessError:
            print("Json contents:")
            print(json.dumps(workflow, indent=4))
            raise

        try:
            run_id = res["run_id"]
        except KeyError:
            print(res)
            raise

    # now we have the run_id
    print(f"Started run with ID {run_id}")
    details = dbjcall(f"runs get --run-id {run_id}")
    print(f"Follow job details at {details['run_page_url']}")

    if out_json:
        json.dump({"run_id": run_id}, out_json)


def discover_and_push_wheels(globpath: str, test_folder: DbfsLocation) -> List[str]:
    result = []
    for item in Path().glob(globpath):
        remote_path = f"{test_folder.remote}/{item.parts[-1]}"
        print(f"pushing {item} to test folder")
        dbfscall(f"cp {item} {remote_path}")
        result.append(remote_path)

    return result


def archive_and_push(test_path: str, test_folder: DbfsLocation, dry_run=False):
    with tempfile.TemporaryDirectory() as tmp:
        test_path = Path(test_path)
        print(f"now archiving {test_path}")
        archive_path = shutil.make_archive(
            str(Path(tmp) / "tests"),
            "zip",
            test_path / "..",
            base_dir=test_path.parts[-1],
        )
        print("now pushing test archive to test folder")

        if not dry_run:
            dbfscall(f"cp {archive_path} {test_folder.remote}/tests.zip")
        else:
            print(
                f"DRY-RUN: Call: dbfs cp {archive_path} {test_folder.remote}/tests.zip"
            )
    return f"{test_folder.local}/tests.zip"


def push_main_file(
    test_folder: DbfsLocation, main_script: IO[str] = None, dry_run=False
) -> DbfsLocation:
    print("now pushing test main file")
    main_file = test_folder / "main.py"
    with tempfile.TemporaryDirectory() as tmp:
        with open(Path(tmp) / "main.py", "w") as f:
            if main_script:
                f.write(main_script.read())
            else:
                print("Using default main script test_main.py")
                f.write(inspect.getsource(test_main))
        if not dry_run:
            dbfscall(f"cp {tmp}/main.py {main_file.remote}")
        else:
            print(f"DRY-RUN: Call: dbfs cp {tmp}/main.py {main_file.remote}")
    return main_file



from collections import namedtuple

_Edge = namedtuple("_Edge", "source target")


class Edge(_Edge):
    """An edge going from source to target.
    Additional details can be added, but will not be considered in equality tests."""

    # The complicated heritage with __new__ and namedtuple ensure that I have to write
    # no further comparison code, I can simply say
    # set(list(Edges from code)) == set(list(Edges from diagram))

    def __new__(cls, source: str, target: str, *, style=None):
        self = super().__new__(cls, source, target)
        self.style = style
        return self



from datetime import date, datetime, timezone

from .TimeSequence import TimeSequence  # noqa: F401


def dt_utc(*args, **kwargs) -> datetime:
    if args or kwargs:
        return datetime(*args, tzinfo=timezone.utc, **kwargs)
    else:
        return datetime.now(timezone.utc)


dt_iso = datetime.fromisoformat
d_iso = date.fromisoformat



import argparse

from spetlrtools.diagrams.DiagramMarkupLibraryParser import DiagramMarkupLibraryParser
from spetlrtools.diagrams.DrawioDiagramParser import DrawioDiagramParser


def process_diagram(
    diagram_path: str,
    code_path: str,
    start_tag: str = DiagramMarkupLibraryParser.yaml_block_start_tag,
    stop_tag: str = DiagramMarkupLibraryParser.yaml_block_end_tag,
    include_files: str = DiagramMarkupLibraryParser.include_files,
) -> int:
    """Collect the set of edges from the diagram and from the code.
    Compare for equality, return the exit code."""
    diagram_edges = set(e.get_Edge() for e in DrawioDiagramParser(diagram_path).parse())

    lib_parser = DiagramMarkupLibraryParser(code_path)
    lib_parser.yaml_block_start_tag = start_tag
    lib_parser.yaml_block_end_tag = stop_tag
    lib_parser.include_files = include_files
    code_edges = set(lib_parser.get_relations())
    if diagram_edges == code_edges:
        return 0

    print("in both:", len(code_edges.intersection(diagram_edges)))
    print("only in diagram:", len(diagram_edges - code_edges))
    for edge in diagram_edges - code_edges:
        print(edge)
    print("only in code:", len(code_edges - diagram_edges))
    for edge in code_edges - diagram_edges:
        print(edge)

    return -1


def main():
    parser = argparse.ArgumentParser(
        description="Compare DrawIO diagram edges to docstrings."
    )

    parser.add_argument(
        "--code-path", help="Root of the codebase to parse.", required=True
    )
    parser.add_argument(
        "--diagram", help="Path to the diagram to process.", required=True
    )
    parser.add_argument(
        "--start-tag",
        help="Tag name for use in docstrings.",
        default=DiagramMarkupLibraryParser.yaml_block_start_tag,
    )
    parser.add_argument(
        "--stop-tag",
        help="Tag name for use in docstrings.",
        default=DiagramMarkupLibraryParser.yaml_block_end_tag,
    )

    parser.add_argument(
        "--include-files",
        help="Regex to determine which files to read.",
        default=DiagramMarkupLibraryParser.include_files,
    )
    args = parser.parse_args()

    ret = process_diagram(
        diagram_path=args.diagram,
        code_path=args.code_path,
        start_tag=args.start_tag,
        stop_tag=args.stop_tag,
        include_files=args.include_files,
    )
    if ret:
        exit(ret)



"""
This is the default main file that is pushed to databricks to launch the test task.
Its tasks are
- to unpack the test archive,
- print a sequence of marker characters to identify the start
  of python executing in the output
- run the tests using pytest
This file is not intended to be used directly.
"""
import argparse
import json
import os
import shutil
import sys
from hashlib import sha1
from pathlib import Path

import pytest

# print a marker message to screen to mark the start of python output
# make it invisible using only "\u200E\u200F\u2060" https://invisible-characters.com/
marker = "".join(list("\u200E\u200F\u2060")[i % 3] for i in sha1(b"my marker").digest())


def test_main():
    """Main function to be called inside the test job task. Do not use directly."""
    parser = argparse.ArgumentParser(description="Run Test Cases.")

    # location to use for current run. Usually the cluster logs base folder
    parser.add_argument("--basedir")

    # relative path of test folder in test archive
    parser.add_argument("--folder")

    # archive of test files to use
    parser.add_argument("--archive")

    # archive of test files to use
    parser.add_argument("--pytestargs")

    args = parser.parse_args()

    extra_args = json.loads(args.pytestargs)

    # the basedir folder should exist because it is also the log destination
    # however we have seen cases where it does not exist yet at the start of the job,
    # so let's create it.
    Path(args.basedir).mkdir(parents=True, exist_ok=True)

    # move to basedir so that simple imports from one test to another work
    os.chdir(args.basedir)
    sys.path = [os.getcwd()] + sys.path

    # unzip test archive to base folder
    shutil.unpack_archive(args.archive)

    # Ensure Spark is initialized before any tests are run
    # the import statement is inside the function so that the outer file
    # can be imported even where pyspark may not be available
    from spetlr.spark import Spark

    Spark.get()

    print()
    print(marker)
    print()

    retcode = pytest.main(["-x", args.folder, *extra_args])
    if retcode.value:
        raise Exception("Pytest failed")


if __name__ == "__main__":
    test_main()



import json
import subprocess
import sys


def db_check():
    """check that the databricks cli is correctly configured
    and has an up-to-date authorization token.
    Returns Nothing.
    Ends the program if this check fails."""
    try:
        import databricks_cli  # noqa
    except ImportError:
        print("Databricks CLI is not installed", file=sys.stderr)
        exit(-1)

    try:
        dbjcall("clusters list  --output=JSON")
    except subprocess.CalledProcessError:
        print(
            "Could not query databricks state. Is your token up to date?",
            file=sys.stderr,
        )
        exit(-1)
    dbcall("jobs configure --version=2.1")


def dbjcall(command: str):
    """Run the command line "databricks "+command and return the unpacked json string."""
    p = subprocess.run(
        "databricks " + command,
        shell=True,
        stdout=subprocess.PIPE,
        check=True,
        text=True,
    )
    try:
        return json.loads(p.stdout)
    except:  # noqa OK because we re-raise
        print(p.stdout)
        raise


def dbfscall(command: str):
    """Run the command line "dbfs "+command and return the output string."""
    p = subprocess.run(
        "dbfs " + command,
        shell=True,
        stdout=subprocess.PIPE,
        check=True,
        text=True,
    )
    return p.stdout


def dbcall(command: str):
    """Run the command line "databricks "+command and return the resulting string."""
    p = subprocess.run(
        "databricks " + command,
        shell=True,
        check=True,
        stdout=subprocess.PIPE,
        text=True,
    )
    return p.stdout



import sys
from http.server import BaseHTTPRequestHandler, HTTPServer
from types import SimpleNamespace
from urllib.parse import parse_qsl, urlparse


class LocalRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        # extract the callback query
        self.server.last_query = SimpleNamespace(
            **dict(parse_qsl(urlparse(self.path).query))
        )

        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(
            bytes(
                """<html><head><title>Deploy Script Databricks Login</title></head>
                <body>
                <h>Deploy Script Databricks Login</h>
                <p>You can close this tab.</p>
                </body></html>
                """,
                "utf-8",
            )
        )

    def log_message(self, format, *args):
        """We don't want the output that someone connected to us,
        because the query contains a secret"""
        print("Callback received.", file=sys.stderr)


class LocalServer(HTTPServer):
    def __init__(self):
        super().__init__(("localhost", 0), LocalRequestHandler)
        self.timeout = 60



import os
import re
from textwrap import dedent
from typing import List

import yaml

from spetlrtools.diagrams.Edge import Edge
from spetlrtools.diagrams.HTMLStripper import condense_whitespace


class DiagramDefinitionError(Exception):
    pass


class DiagramMarkupLibraryParser:
    """Go through a library of files, look for tagged blocks
    and extract the configuration. Based on the configuration,
    return the desired set of diagram edges."""

    yaml_block_start_tag = "```diagram"
    yaml_block_end_tag = "```"
    include_files = ".py$"

    def __init__(self, package_path: str):
        self.package_path = package_path

    def _get_all_blocks(self):
        """generator to get all blocks of code enclosed by tags"""
        pattern = re.compile(
            f"(?<={self.yaml_block_start_tag})"  # lookbehind assertion
            ".*?"  # any code whatsoever, but never include the end tag.
            f"(?={self.yaml_block_end_tag})",  # lookahead assertion
            flags=re.DOTALL,
        )
        for root, dirs, files in os.walk(self.package_path):
            for file in files:
                if re.search(self.include_files, file):
                    file_path = os.path.join(root, file)
                    with open(file_path, encoding="utf-8") as f:
                        # match the tag sequence absolutely anywhere in the file
                        # this is usually in docstrings.
                        for match in pattern.findall(f.read()):
                            yield file_path, str(match)

    def _build_config(self):
        """Parse all the code blocks to build the aggregated
        YAML configuration across the library"""
        config = {}
        for file_path, block in self._get_all_blocks():
            try:
                obj = yaml.safe_load(dedent(block))
            except yaml.YAMLError:
                print(f"WARNING: Skipping unparsable yaml in {file_path}")
                continue
            if not isinstance(obj, dict):
                print(f"WARNING: block in {file_path} is not a dict")
                continue
            for key, node in obj.items():
                if not isinstance(node, dict):
                    print(f"WARNING: The contents of {key} are not a dict: {node}.")
                    print("         If you want a simple alias use:")
                    print(f"           {key}:")
                    print(f"           name: {node}")
                    continue

                try:
                    node["name"] = condense_whitespace(node["name"])
                except KeyError:
                    print(f"WARNING: Missing name in node {key} of {file_path}")
                    continue

                if key in config:
                    if config[key] != node:
                        raise DiagramDefinitionError(
                            f"Node {key} in {file_path} "
                            "conflicts with earlier definitions"
                        )
                config[key] = node
        return config

    def get_relations(self) -> List[Edge]:
        """Returns all edges as requested in the library configuration."""
        config = self._build_config()

        def resolve_name(key):
            if key in config:
                return config[key]["name"]
            else:
                return key

        for key, node in config.items():
            nodename = node["name"]  # presenece already ensured.
            for source in node.get("incoming", []):
                if isinstance(source, str):
                    # incoming edges go *to* the nodename
                    yield Edge(resolve_name(source), nodename)
                else:
                    raise NotImplementedError("styles not supported yet")
            for target in node.get("outgoing", []):
                if isinstance(target, str):
                    # outgoing edges go *from* the nodename
                    yield Edge(nodename, resolve_name(target))
                else:
                    raise NotImplementedError("styles not supported yet")



import base64
import xml.etree.ElementTree as ET
import zlib
from itertools import chain
from typing import Dict, List, Optional
from urllib.parse import unquote

from spetlrtools.diagrams.Edge import Edge
from spetlrtools.diagrams.HTMLStripper import HTMLStripper, condense_whitespace


class DiagramNode:
    """A node in a drawio diagram"""

    def __init__(self, details):
        self.id = details["id"]
        self.label = ""
        for key in ["value", "label"]:
            try:
                self.label = HTMLStripper.strip(details[key])
                break
            except KeyError:
                continue


class DiagramEdge(DiagramNode):
    """An edge linking two diagram nodes, using diagram specific identifiers."""

    def __init__(self, details):
        super().__init__(details)

        self.source_id = details["source"]
        self.target_id = details["target"]
        self.source_node: Optional[DiagramNode] = None
        self.target_node: Optional[DiagramNode] = None

    @property
    def source_label(self):
        return self.source_node.label if self.source_node else self.source_id

    @property
    def target_label(self):
        return self.target_node.label if self.target_node else self.target_id

    def __str__(self):
        return f"< {self.source_label} -- {self.target_label} >"

    def get_Edge(self) -> Edge:
        return Edge(
            condense_whitespace(self.source_label),
            condense_whitespace(self.target_label),
        )


class DrawioDiagramParser:
    """This class parses the diagram and returns all found edges."""

    def __init__(self, path):
        self.path = path
        self.edges: List[DiagramEdge] = []
        self.nodes: Dict[str, DiagramNode] = {}

    def _get_contents(self, path: str):
        """Get the raw document contents even for 'editable png' and 'editable svg'."""
        if path.endswith("png"):
            try:
                from PIL import Image
            except ImportError:
                raise Exception("You need to install 'pillow' to work with png files.")

            im = Image.open(path)
            im.load()
            conts = unquote(im.info["mxfile"])
            return conts
        elif path.endswith(".svg"):
            with open(path, "r", encoding="utf-8") as f:
                conts = f.read()
                svg = ET.fromstring(conts)

            return svg.attrib["content"]
        else:
            # both .drawio and .xml fall into this case
            with open(path, "r", encoding="utf-8") as f:
                return f.read()

    def _deflate_nodes(self, et: ET) -> ET:
        """If the diagram is saved as compressed, decompress it."""
        for diagram in et.iter("diagram"):
            if len(diagram):
                # d has nested xml nodes
                continue
            b64 = base64.b64decode(diagram.text)
            full = unquote(zlib.decompress(b64, -15).decode("utf-8"))
            diagram.text = ""
            diagram.insert(0, ET.fromstring(full))
        return et

    def parse(self) -> List[DiagramEdge]:
        """Read the diagram and return the edges."""
        conts = self._get_contents(self.path)
        et = ET.fromstring(conts)
        et = self._deflate_nodes(et)

        _edges = []

        for cell in chain(et.iter("mxCell"), et.iter("object")):
            try:
                node = DiagramNode(cell.attrib)
            except KeyError:
                continue
            self.nodes[node.id] = node

            try:
                _edges.append(DiagramEdge(cell.attrib))
            except KeyError:
                # the node is not a useful edge
                continue

        for e in _edges:
            e.source_node = self.nodes[e.source_id]
            e.target_node = self.nodes[e.target_id]
            if e.source_node.label and e.target_node.label:
                self.edges.append(e)

        return self.edges



import warnings
from itertools import chain
from typing import List

import pyspark.sql.functions as f
from pyspark.sql import DataFrame


def get_difference_between_two_dfs(
    df1: DataFrame,
    df2: DataFrame,
    join_cols: List[str],
    ignore_cols: List[str] = None,
    print_result: bool = False,
) -> DataFrame:
    """
    This function compares two dataframe using hashing of the columns.
    The function uses left join, which that all records from df1 are matched with records from df2.

    NB:
    The function only compares columns that appear in both tables.

    :param df1: A dataframe to compared with df2
    :param df2: A dataframe to compared with df1
    :param join_cols: A string list for the join column names
    :param ignore_cols: Columns to be ignored when comparing
    :param print_result: If true, the differences is printed

    :return: A dataframe with the differences between df1 and df2
            The column names are prefixed with df1_ or df_2 for easier comparison

    | join_col| df1_col1 | df2_col1 |
    |---------|----------|----------|
    | 1       | x        | y        |
    |---------|----------|----------|

    """

    # Lowering join and ignore columns
    join_cols = [col.lower() for col in join_cols]

    if ignore_cols is None:
        ignore_cols = []
    else:
        ignore_cols = [col.lower() for col in ignore_cols]

    # Get columns without the ignored columns
    df1_cols_without_ignores = set(
        [col.lower() for col in df1.columns if col.lower() not in ignore_cols]
    )

    df2_cols_without_ignores = set(
        [col.lower() for col in df2.columns if col.lower() not in ignore_cols]
    )

    if not df2_cols_without_ignores.issubset(df1_cols_without_ignores):
        warnings.warn(
            "Some of the columns in df1 is not in df2. The tables will therefore always be different!\n"
            "Remember, this function only compares the columns that appear in both tables.\n"
            "Continuing..."
        )

    if not df1_cols_without_ignores.issubset(df2_cols_without_ignores):
        warnings.warn(
            "Some of the columns in df2 is not in df1. The tables will therefore always be different!\n"
            "Remember, this function only compares the columns that appear in both tables. \n"
            "Continuing..."
        )

    # Select same columns that appear in both tables. Without the ignore_cols.
    _cols = list(df1_cols_without_ignores & df2_cols_without_ignores)

    # Generate hash columns
    df1_sub = df1.select(_cols).withColumn(
        "row_sha2", f.sha2(f.concat_ws("||", *_cols), 256)
    )
    df2_sub = df2.select(_cols).withColumn(
        "row_sha2", f.sha2(f.concat_ws("||", *_cols), 256)
    )

    # Select the order as a.col1, b.col1, a.col2, b.col2.... this way it is easier to see differences.
    # Also prefixes df1_ and df2_

    order_list = [
        [f.col(f"a.{col}").alias(f"df1_{col}"), f.col(f"b.{col}").alias(f"df2_{col}")]
        for col in _cols
        if col not in ignore_cols + join_cols
    ]

    order_list = list(chain.from_iterable(order_list))

    # Generating the differences
    _join_cols_a = [f"a.{col}" for col in join_cols]

    df_res = (
        df1_sub.alias("a")
        .join(df2_sub.alias("b"), join_cols, "left")
        .filter("a.row_sha2 != b.row_sha2")
        .select(*_join_cols_a, *order_list)
    )

    if print_result:
        print(f"The number of records with differences: {df_res.count()}")
        df_res.show()

    return df_res



from dataclasses import dataclass


@dataclass
class DbfsLocation:
    """A Dbfs location needs to be accessed through dbfs:/ when acting remotely
    through the cli.
    When running in the cluster, the files are accessed locally under /dbfs/.
    This class unifies the two methods under one class
    """

    base: str = ""

    @property
    def remote(self):
        return "dbfs:/" + self.base

    @property
    def local(self):
        return "/dbfs/" + self.base

    def __truediv__(self, value) -> "DbfsLocation":
        return DbfsLocation(base=self.base + "/" + value)

    @classmethod
    def from_str(cls, val: str) -> "DbfsLocation":
        if val[:6] not in ["dbfs:/", "/dbfs/"]:
            raise ValueError("Invalid dbfs path")

        return cls(base=val[:6])

    @remote.setter
    def set_remote(self, remote: str) -> None:
        if not remote.startswith("dbfs:/"):
            raise AssertionError("Remote locations must start with dbfs:/")
        self.base = remote[6:]

    @local.setter
    def set_local(self, local: str) -> None:
        if not local.startswith("/dbfs/"):
            raise AssertionError("Local locations must start with /dbfs/")
        self.base = local[6:]



import re
import subprocess
import tempfile
import time

from spetlrtools.test_job import test_main
from spetlrtools.test_job.dbcli import dbcall, dbfscall, dbjcall


class DbfsFileDoesNotExist(Exception):
    pass


class RunDetails:
    """Object representing the details of a job run"""

    details: dict

    def __init__(self, run_id: int):
        self.run_id = run_id
        self.refresh()
        print(f"Job details: {self.details['run_page_url']}")

    def refresh(self):
        """refresh the internal details by querying databricks"""
        self.details = dbjcall(f"runs get --run-id {self.run_id}")

    def cancel(self):
        """Cancel the run on databricks."""
        dbcall(f"runs cancel --run-id {self.run_id}")

    def get_stdout(self, task_key: str) -> str:
        """Return the driver stdout from the cluster logs."""
        print(f"Getting stdout for {task_key}")
        (task,) = [t for t in self.details["tasks"] if t["task_key"] == task_key]
        log_destination = task["new_cluster"]["cluster_log_conf"]["dbfs"]["destination"]
        cluster_instance = task["cluster_instance"]["cluster_id"]
        print(log_destination, cluster_instance)
        log_stout_path = f"{log_destination}/{cluster_instance}/driver/stdout"
        with tempfile.TemporaryDirectory() as tmp:
            try:
                self.wait_until_exists(log_stout_path)
            except DbfsFileDoesNotExist:
                return f"-=ERROR FETCHING LOG FILES FROM  {log_stout_path}=-"

            dbfscall(f"cp {log_stout_path} {tmp}/stdout")
            with open(f"{tmp}/stdout", encoding="utf-8") as f:
                return clean_cluster_output(f.read())

    def wait_until_exists(self, location: str, tries=20, sleep_s=1):
        """Wait until the file exists on dbfs."""
        # sometimes cluster logs do not appear immediately.
        # TODO: expose the wait parameters on the cli
        for i in range(tries):
            try:
                dbfscall(f"ls {location}")
                return
            except subprocess.CalledProcessError:
                time.sleep(sleep_s)

        raise DbfsFileDoesNotExist()


def clean_cluster_output(raw_stdout: str) -> str:
    """
    Clean in two steps:
    - Step 1 is to find the sequence of invisible marker characters
      and discard everything before that.
    - Step 2 is to use regexp to get rid of interspersed java warnings.
    """
    # Step 1 split on marker
    try:
        _, raw_stdout = raw_stdout.split(test_main.marker)
    except ValueError:
        # no marker in input?! try moving on regardless
        pass

    # step 2 remove GC warnings
    timestamp = (
        r"\d{4}-\d{2}-\d{2}"  # Year, month, day
        r"T\d{2}:\d{2}:\d{2}"  # Hour, minute, second
        r"\.\d{3}"  # millseconds
        r"\d*"  # further fractional seconds
        r"(\+\d+|-\d+|Z)"  # Time zone info
    )
    pattern = (
        r"^"  # preceeded by a newline or start of string
        f"({timestamp}:? )+"  # At least 1 timestamp followed by ": "
        r"(\d+\.\d+: )?"  # unknown indicator
        r"\["  # any block starting with a [
        r".+$[\r\n]*"  # everything to end of line and all following newlines
    ).replace(" ", r"\s*")

    pat = re.compile(pattern, flags=re.MULTILINE)

    return pat.sub("", raw_stdout)



"""
- find the test job (by ID or discover by tag)
- get job status:
- tasks: pending: 0 running: 0 success: 0 failed: 0
- any time a task finishes, print the log
- add a fail_fast so that any time a task fails, job is cancelled
"""
import argparse
import json
import sys
import time
from collections import defaultdict
from dataclasses import dataclass
from typing import IO, List, Optional

from spetlrtools.test_job.dbcli import db_check
from spetlrtools.test_job.RunDetails import RunDetails


def setup_fetch_parser(subparsers):
    """
    Adds a subparser for the command 'fetch'.
    :param subparsers: must be the object returned by ArgumentParser().add_subparsers()
    :return:
    """
    parser: argparse.ArgumentParser = subparsers.add_parser(
        "fetch", description="Return test run result."
    )
    parser.set_defaults(func=fetch_main)

    # cluster argument pair
    runid_config = parser.add_mutually_exclusive_group(required=True)
    runid_config.add_argument(
        "--runid",
        type=int,
        help="Run ID of the test job",
        default=None,
    )
    runid_config.add_argument(
        "--runid-json",
        type=argparse.FileType("r"),
        help="File with JSON document describing the Run ID of the test job.",
        default=None,
    )

    parser.add_argument(
        "--stdout",
        type=argparse.FileType("w"),
        required=False,
        help="Output test stdout to this file.",
        default=None,
    )

    parser.add_argument(
        "--failfast",
        action="store_true",
        help="Stop and cancel job on first failed task.",
    )


def collect_args(args):
    """Post process the arguments of the ."""
    if args.runid is None:
        args.runid = json.load(args.runid_json)["run_id"]

    return args


def fetch_main(args):
    """
    Main function of the 'fetch' command. Only to be used via the cli.
    :param args: the parsed arguments from the fetch subparser
    :return:
    """
    db_check()

    # Post process the arguments
    if args.runid is None:
        args.runid = json.load(args.runid_json)["run_id"]

    if fetch(args.runid, args.stdout, args.failfast):
        print("Run failed")
        sys.exit(-1)


@dataclass
class TaskState:
    """The result state of any workflow or task"""

    task_key: str
    life_cycle_state: str
    result_state: str
    end_time: int

    @property
    def ended(self):
        """Has the workflow or task ended?"""
        return self.end_time != 0

    @property
    def result(self):
        """String representing the state of the workflow or task."""
        if not self.ended:
            return self.life_cycle_state
        else:
            return self.result_state

    @property
    def success(self):
        """Has the workflow or task succeeded?"""
        return self.result_state.upper() == "SUCCESS"

    @classmethod
    def fromJson(cls, jobj):
        """Create the result state of the workflow or task from the json object
        returned by the databricks api."""
        state = jobj["state"]
        return cls(
            task_key=jobj["task_key"] if "task_key" in jobj else jobj["run_name"],
            life_cycle_state=state["life_cycle_state"],
            result_state=state["result_state"] if "result_state" in state else "",
            end_time=jobj["end_time"] if "end_time" in jobj else 0,
        )


@dataclass
class MultiTaskState:
    """Result state of a multiTask workflow"""

    overall: Optional[TaskState]
    tasks: List[TaskState]

    @classmethod
    def fromJson(cls, jobj):
        """Create the Result state of a multiTask workflow from the json object returned
        by the databricks api."""

        return cls(
            overall=TaskState.fromJson(jobj),
            tasks=[TaskState.fromJson(task) for task in jobj["tasks"]],
        )

    def accumulate(self):
        """return a dictionary of {'STATE STRING':counts} where counts represents the
        number of sub-tasks that share the same run state."""
        counts = defaultdict(int)
        for task in self.tasks:
            counts[task.result] += 1
        return counts

    def print_status(self):
        """Print the overall result state represented by this object."""
        print(
            "Overall state:",
            self.overall.result,
            "| Task states:",
            " | ".join(f"{k}: {v}" for k, v in self.accumulate().items()),
        )


def fetch(run_id: int, stdout_file: IO[str] = None, failfast=False):
    """Fetch main function.
    See the cli help for parameter descriptions and functionality.
    Can be used programmatically."""
    run = RunDetails(run_id)
    last_state = None
    stdouts = {}
    while True:
        state = MultiTaskState.fromJson(run.details)
        if last_state is None or state != last_state:
            last_state = state
            state.print_status()

        if failfast:
            if any(task.ended and not task.success for task in state.tasks):
                break

        for task in state.tasks:
            if task.ended and task.task_key not in stdouts:
                out = run.get_stdout(task.task_key)
                if stdout_file is None:
                    print(out)
                stdouts[task.task_key] = out

        if state.overall.ended:
            break
        time.sleep(5)
        run.refresh()

    if stdout_file is not None:
        for k, v in stdouts.items():
            stdout_file.write("=" * 50 + f"\nTask Output from {k}\n" + "=" * 50 + "\n")
            stdout_file.write(v)

    if last_state.overall.success:
        print("Run result SUCCESS!")
        return 0
    else:
        if not (last_state.overall.ended):
            print("A task failed. Cancelling test run.")
            run.cancel()
        return 1
